/* Copyright (C) 2008 Mig Internacional
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * */


#include <linux/init.h>
#include <linux/kernel.h>	/* printk(), min() */
#include <linux/module.h> /* Standard on every linux module */
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/dma.h> // dma_ext3, EXTDMA3_RX_DMA_NBR, DMA_VERBOSE_ON_ERROR
#include <asm/arch/dma.h>
#include <asm/uaccess.h>	/* copy_*_user */
#include <linux/cdev.h> /* module initialization */
#include <linux/interrupt.h>
#include <asm/arch/hwregs/intr_vect.h> // EXT_INTR_VECT
#include <asm/arch/hwregs/bif_slave_defs.h> //access to the bif_slave register, REG_WR, rw_arb_cfg
#include <asm/arch/hwregs/reg_map.h> //regi_bif_slave
#include <asm/io.h> //PAGE_SIZE
#include  <linux/dma-mapping.h>

#include <linux/types.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/pci.h>

#define MEM_CSP4_START (0x20000000)
#define MEM_NON_CACHEABLE (0x80000000)
#define FPGA_MAJOR 42
#define FPGA_MINOR 0
#define REQ_DMA 0x83
#define ACK_INT 0x82
#define REQ_INT 0x81
#define READ_REG 0x80
#define FPGA_DMA_CHANNEL 0x0
#define DMA_BUF_SIZE 76800

static DECLARE_WAIT_QUEUE_HEAD(frame_wq);
volatile unsigned long *fpga;
unsigned int bif_slave = 1;
static int frame_ready = 0;
struct frame_reader_data * fdata;
static unsigned long *fpga_dma_buf;
static dma_addr_t dmabuf_phys_addr;

// taken from arch/cris/drivers/pci/dma.c
struct dma_coherent_mem {
	void		*virt_base;
	u32		device_base;
	int		size;
	int		flags;
	unsigned long	*bitmap;
};

// taken from arch/cris/drivers/pci/dma.c
void *my_dma_alloc_coherent(struct device *dev, size_t size,
			   dma_addr_t *dma_handle, gfp_t gfp) {
	void *ret;
	struct dma_coherent_mem *mem = dev ? dev->dma_mem : NULL;
	int order = get_order(size);
	/* ignore region specifiers */
	gfp &= ~(__GFP_DMA | __GFP_HIGHMEM);

	if (mem) {
		int page = bitmap_find_free_region(mem->bitmap, mem->size,
						     order);
		if (page >= 0) {
			*dma_handle = mem->device_base + (page << PAGE_SHIFT);
			ret = mem->virt_base + (page << PAGE_SHIFT);
			memset(ret, 0, size);
			return ret;
		}
		if (mem->flags & DMA_MEMORY_EXCLUSIVE)
			return NULL;
	}

	if (dev == NULL || (dev->coherent_dma_mask < 0xffffffff))
		gfp |= GFP_DMA;

    ret = (void *)__get_free_pages(gfp, order);

	if (ret != NULL) {
		memset(ret, 0, size);
		*dma_handle = virt_to_phys(ret);
	}
	return ret;
}


// taken from arch/cris/drivers/pci/dma.c
void my_dma_free_coherent(struct device *dev, size_t size,
			 void *vaddr, dma_addr_t dma_handle)
{
	struct dma_coherent_mem *mem = dev ? dev->dma_mem : NULL;
	int order = get_order(size);
	
	if (mem && vaddr >= mem->virt_base && vaddr < (mem->virt_base + (mem->size << PAGE_SHIFT))) {
		int page = (vaddr - mem->virt_base) >> PAGE_SHIFT;

		bitmap_release_region(mem->bitmap, page, order);
	} else
		free_pages((unsigned long)vaddr, order);
}

struct frame_reader_data {
    int frame_idx;
    struct cdev * cdev;
};

int frame_reader_open(struct inode *inode, struct file *filp) {
    printk("Opening cosa\n");
    return 0;
}

int frame_reader_release(struct inode *inode, struct file *filp) {
    printk("Closing cosa\n");
    frame_ready = 0;
    fdata->frame_idx = 0;
	return 0;
}

ssize_t frame_reader_read(struct file *filp, char __user *buf, size_t count,
                          loff_t *f_pos) {
    if (!frame_ready) {
        printk("about to ask frame\n");
        fpga[0x84] = dmabuf_phys_addr;
        printk("asked for new frame\n");
        wait_event_interruptible(frame_wq, frame_ready); // fpga_interrupt will awake us
    }
    if (fdata->frame_idx == 4 * DMA_BUF_SIZE) {
        frame_ready = 0; // we just ate it, or fed it to userspace, or whatever
        fdata->frame_idx = 0;
        return 0; //EOF 
    }
    if (fdata->frame_idx + count > 4 * DMA_BUF_SIZE) {
        count = 4 * DMA_BUF_SIZE - fdata->frame_idx;
    }
    copy_to_user(buf, ((char*) fpga_dma_buf)+fdata->frame_idx, count);
    fdata->frame_idx += count;
    return count;
}

static ssize_t frame_reader_write(struct file * file, const char * buf, size_t count, 
                           loff_t *off) {
    return count; //success
}

irqreturn_t fpga_interrupt(int irq, void *dev_id) {
    fpga[ACK_INT] = 0x01; // ack interrupt
    printk("int ack\n");
    fdata->frame_idx = 0;
    frame_ready = 1;
    wake_up_interruptible(&frame_wq);
    return IRQ_HANDLED;
}

static void frame_reader_exit(void) {
    free_irq(EXT_INTR_VECT, NULL);
    unregister_chrdev_region(MKDEV(FPGA_MAJOR, FPGA_MINOR), 1);
    cdev_del(fdata->cdev);
    kfree(fdata);
    iounmap(fpga);
    my_dma_free_coherent(NULL, DMA_BUF_SIZE, fpga_dma_buf, dmabuf_phys_addr);
    printk(KERN_ALERT "Module frame_reader unloaded\n");
}

struct file_operations frame_reader_fops = {
    .owner = THIS_MODULE,
    .llseek = NULL,
    .read = frame_reader_read,
    .write = frame_reader_write,
    .ioctl = NULL,
    .open = frame_reader_open,
    .release = frame_reader_release,
};

static int frame_reader_init(void) {
    int result;
    fpga = ioremap(MEM_CSP4_START|MEM_NON_CACHEABLE, 8192);
    result = register_chrdev_region(MKDEV(FPGA_MAJOR, FPGA_MINOR), 1, "frame_reader");
    if (result < 0) {
        printk("Unable to get device number. Error code: %d\n", result);
        return result;
    }
    fdata = kmalloc(sizeof(struct frame_reader_data), GFP_KERNEL);
    if (fdata == NULL) {
        printk("Unable to get memory for inner data structure\n");
        unregister_chrdev_region(MKDEV(FPGA_MAJOR, FPGA_MINOR), 1);
        return -ENOMEM;
    }
	memset(fdata, 0, sizeof(struct frame_reader_data));
    fdata->cdev = cdev_alloc();
    fdata->cdev->ops = &frame_reader_fops;
    fdata->cdev->owner = THIS_MODULE;
    cdev_add(fdata->cdev, MKDEV(FPGA_MAJOR, FPGA_MINOR), 1);
    if(request_irq(EXT_INTR_VECT, fpga_interrupt, SA_INTERRUPT,
                   "FPGA Interrupt", NULL)) {
        printk(KERN_ERR "Can't allocate FPGA interruptions\n");
        cdev_del(fdata->cdev);
        unregister_chrdev_region(MKDEV(FPGA_MAJOR, FPGA_MINOR), 1);
        kfree(fdata);
        return -EBUSY;
    }
    reg_bif_slave_rw_arb_cfg slave_reg = REG_RD(bif_slave, regi_bif_slave, 
                                                rw_arb_cfg);
    slave_reg.release = 2; // So the CPU shares the bus with other devices
    REG_WR(bif_slave, regi_bif_slave, rw_arb_cfg, slave_reg);
    printk(KERN_ALERT "Module frame_reader loaded\n");
    fpga_dma_buf = my_dma_alloc_coherent(NULL, DMA_BUF_SIZE, &dmabuf_phys_addr, 
            GFP_KERNEL);
    if (!fpga_dma_buf) {
        printk("Could not get DMA buffer for FPGA\n");
        free_irq(EXT_INTR_VECT, NULL);
        cdev_del(fdata->cdev);
        unregister_chrdev_region(MKDEV(FPGA_MAJOR, FPGA_MINOR), 1);
        kfree(fdata);
        return -ENOMEM;
    }
    return 0;
}

module_init(frame_reader_init);
module_exit(frame_reader_exit);
