`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 		 MIG Internacional
// Engineer: 		 Juan Jos Rondn Ruiz
// 
// Create Date:    10:56:58 07/18/2008 
// Design Name: 
// Module Name:    trackwrite 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module trackwrite(sys_sdclk,		//Clock signal input from SDRAM control (Not Used)
						sys_sdwe,		//WE signal for SDRAM
						sys_sdcas,		//CAS signal for SDRAM
						sys_sdras,		//RAS signal for SDRAM
						addr,				//Address bus
						data,				//Data bus
						ba0,				//Bank select 0 for SDRAM
						ba1,				//Bank select 1 for SDRAM
						oe,				//RD Signal from Processor
						ce1,				//CSP4 Signal from Processor
						irq,				//Interrupt Signal from Processor
						brout,			//Bus Request Output
						sys_busen,		//Enable Signals for SDRAM
						done,				//Done request data from Processor (Debugg Signal)
						brin,				//Bus Request Input
						clk,				//Clock Signal (100 MHz)
						flag,				//FPGA master of bus indicator (Debugg Signal)
						bg,				//Bus Grant Signal
						sys_sdclkout,	//Clock signal output for SDRAM
						writeok,			//Done write data in SDRAM indicator (Debugg Signal)
						we);				//WR3 Signal from Processor

input clk,sys_sdclk,oe,ce1,brin,we;
inout sys_sdwe,sys_sdcas,sys_sdras,ba0,ba1,bg;
inout [12:0] addr;
inout [31:0] data;
output irq,brout,sys_busen,done,sys_sdclkout,flag,writeok;

reg irq=1;
reg done=0;
reg ctrdata1=0; 						//Data bus output control (for read instruction from Processor)
wire ctrdata2; 						//Data bus output control (for write data in SDRAM)
reg [31:0] dataout; 					//Data bus output register
wire [31:0] dataout2; 				//Data bus output for write data in SDRAM
reg [31:0] dirram=32'hFFFFFFFF; 	//Initial SDRAM address
wire clkin;								//Clock input after buffer
reg doneirq=0;							//Interrupt is acknowledge by the processor 

IBUF clkinputbuf (.O(clkin),.I(clk)); // bufer input for CLK signal

//Control of Data bus output
always @ (ctrdata1 or ctrdata2 or dataout2 or dirram) begin 
	if (ctrdata2)
		dataout<=dataout2;
	else if (ctrdata1)
		dataout<=dirram;
	else
		dataout<=32'bz;
end
assign data=dataout;

wire write;

assign write = ce1 | we;

// A rising edge of write is end of processor write cycle  
always @ (posedge write) begin
	if (addr[7:0]==8'h84) begin 	//Write data in SDRAM Request
		dirram <= data;				//Read the SDRAM Address
		done <= 1;						//Enable Bus request and Write SDRAM Process
		doneirq <= 0;
	end 
	if (addr[7:0]==8'h82) begin	//Interrupt is acknowledge
		doneirq <= 1;					
		done <= 0;
	end
end
	
//Verify Write data to Processor request
always @ (ce1 or oe or addr) begin
		if (ce1==0) begin 
			if (oe==0) begin 
				if (addr[7:0] == 8'h80) begin 
					ctrdata1 <= 1; 	//Write data to Processor
				end
				else begin
					ctrdata1 <= 0; 
				end
			end
			else begin
				ctrdata1 <= 0;
			end
		end
		else begin
			ctrdata1 <= 0;
		end
end

//States definitions
parameter [2:0] init=0;
parameter [2:0] busreq1=1;
parameter [2:0] busreq2=2;
parameter [2:0] busreq3=3;
parameter [2:0] busreq4=4;
parameter [2:0] busreq5=5;

reg brout=0;
reg ctrbg=0;			//Output control of BG signal
reg bgout=0;
reg [2:0] delay=0; 	//Cycles delay from bus arbitration protocol
reg [2:0] state=init;//State Machine
reg ctrbus=0;			//This signal is high when FPGA have the control of bus
reg flag=0;
wire donedato;

assign bg=ctrbg?bgout:1'bz;

always @ (negedge clkin) begin
	case (state) 
		init: begin					//Initial state
			irq <= 1;				//Dont interrupt the processor
			ctrbg <= 0;				//BG output is high impedance
			brout <= 0; 			//Bus request output is inactive
			if (done) begin		//The Processor require that FPGA write in SDRAM
				brout <= 1; 		//Bus request output is active
				state <= busreq1;	//Next state
				flag <= 0;
				delay <= 0;
				irq <= 1;
			end
		end
		busreq1: begin
			if (bg) begin			//The Processor Grant the bus control to FPGA
				if (delay==3) begin //Wait 
					delay <= 0;
					bgout <= 0;			//Down bgout signal
					ctrbg <= 1;			//The FPGA drive this signal
					state <= busreq2;	
					flag <= 1;
				end
				else
					delay <= delay + 1;
			end
		end
		busreq2: begin
			if (delay==7) begin	//Wait
				delay <= 0;
				state <= busreq3;
			end
			else 
				delay <= delay + 1;
		end
		busreq3: begin
			state <= busreq4;
		end
		busreq4: begin
			ctrbus <= 1;				//Enable Write SDRAM Data
			brout <= 0;
			if (donedato) begin		//Wait until write SDRAM data is done
				ctrbus <= 0;			//Disable write SDRAM data
				irq <= 0;				//Interrupt the processor
				state <= busreq5;
			end
		end
		busreq5: begin
			if (brin) begin			//The CPU require the bus control
				bgout <= 1;				//The FPGA grant the bus control
				ctrbg <= 0;
			end
			if (doneirq) begin		//The CPU acnowledge the interrupt
				irq <= 1;				//Disable interrupt
				state <= init;			//Begin again
			end
		end
		default:
			state <= init;
	endcase
end

wire [12:0] addrout;
wire sys_busen;
wire sys_sdcasout;
wire sys_sdrasout;
wire sys_sdweout;
wire ba0out;
wire ba1out;
wire writeok;
wire ctraddr;
reg clkdiv=1;

//Output control of SDRAM control signals

assign sys_sdcas=ctrbus?sys_sdcasout:1'bz;
assign sys_sdras=ctrbus?sys_sdrasout:1'bz;
assign sys_sdwe=ctrbus?sys_sdweout:1'bz;
assign ba0=ctrbus?ba0out:1'bz;
assign ba1=ctrbus?ba1out:1'bz; 
assign sys_sdclkout=ctrbus?clkdiv:1'bz;
assign addr=ctraddr?addrout:13'bz;

//Clock divisor
always @ (posedge clkin)
	clkdiv <= ~clkdiv;

//This module write data in SDRAM

writedata wrd0 (ctrbus,ctraddr,ctrdata2,donedato,addrout,sys_busen,sys_sdcasout,sys_sdrasout,sys_sdweout,ba0out,ba1out,writeok,dirram,clkdiv,dataout2);


endmodule
